@@ -7,59 +7,63 @@
 --
 -- Nothing monadic in here!
 --
--- (c) The University of Glasgow 2004-2006
---
 -----------------------------------------------------------------------------
 
+{-# LANGUAGE RecordWildCards #-}
+{-# OPTIONS -fno-warn-tabs #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and
+-- detab the module (please do the detabbing in a separate patch). See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#TabsvsSpaces
+-- for details
+
 module StgCmmClosure (
-        SMRep, 
-	DynTag,  tagForCon, isSmallFamily,
+        DynTag,  tagForCon, isSmallFamily,
 	ConTagZ, dataConTagZ,
 
-	ArgDescr(..), Liveness(..), 
-	C_SRT(..), needsSRT,
-
-	isVoidRep, isGcPtrRep, addIdReps, addArgReps,
+        isVoidRep, isGcPtrRep, addIdReps, addArgReps,
 	argPrimRep, 
 
-	LambdaFormInfo,		-- Abstract
+        -- * LambdaFormInfo
+        LambdaFormInfo,         -- Abstract
 	StandardFormInfo,	-- ...ditto...
 	mkLFThunk, mkLFReEntrant, mkConLFInfo, mkSelectorLFInfo,
 	mkApLFInfo, mkLFImported, mkLFArgument, mkLFLetNoEscape,
-	lfDynTag,
+        mkLFBlackHole,
+        lfDynTag,
+        maybeIsLFCon, isLFThunk, isLFReEntrant, lfUpdatable,
 
+        nodeMustPointToIt,
+        CallMethod(..), getCallMethod,
+
+        isKnownFun, funTag, tagForArity,
+
+        -- * ClosureInfo
 	ClosureInfo,
-	mkClosureInfo, mkConInfo, maybeIsLFCon,
+        mkClosureInfo,
+        mkCmmInfo,
 
-	closureSize, closureNonHdrSize,
-	closureGoodStuffSize, closurePtrsSize,
-	slopSize, 
+        -- ** Inspection
+        closureLFInfo, closureName,
 
-	closureName, infoTableLabelFromCI, entryLabelFromCI,
-	closureLabelFromCI,
-	closureTypeInfo,
-	closureLFInfo, isLFThunk,closureSMRep, closureUpdReqd, 
-	closureNeedsUpdSpace, closureIsThunk,
-	closureSingleEntry, closureReEntrant, isConstrClosure_maybe,
-	closureFunInfo,	isStandardFormThunk, isKnownFun,
-        funTag, tagForArity, 
+        -- ** Labels
+        -- These just need the info table label
+        closureInfoLabel, staticClosureLabel,
+        closureRednCountsLabel, closureSlowEntryLabel, closureLocalEntryLabel,
 
-	enterIdLabel, enterLocalIdLabel, 
+        -- ** Predicates
+        -- These are really just functions on LambdaFormInfo
+        closureUpdReqd, closureSingleEntry,
+        closureReEntrant, closureFunInfo,
+        isToplevClosure,
 
-	nodeMustPointToIt, 
-	CallMethod(..), getCallMethod,
+        blackHoleOnEntry,  -- Needs LambdaFormInfo and SMRep
+        isStaticClosure,   -- Needs SMPre
 
-	blackHoleOnEntry,
-
-	getClosureType,
-
-	isToplevClosure,
-	closureValDescr, closureTypeDescr,	-- profiling
-
-	isStaticClosure,
-	cafBlackHoleClosureInfo, 
-
-	staticClosureNeedsLink, clHasCafRefs 
+        -- * InfoTables
+        mkDataConInfoTable,
+        cafBlackHoleInfoTable,
+        staticClosureNeedsLink,
     ) where
 
 #include "../includes/MachDeps.h"
@@ -67,13 +71,9 @@ module StgCmmClosure (
 #define FAST_STRING_NOT_NEEDED
 #include "HsVersions.h"
 
-import ClosureInfo (ArgDescr(..), C_SRT(..), Liveness(..))
-	-- XXX temporary becuase FunInfo needs this one
-
 import StgSyn
 import SMRep
-import CmmDecl ( ClosureTypeInfo(..), ConstrDescription )
-import CmmExpr
+import Cmm
 
 import CLabel
 import StaticFlags
@@ -87,6 +87,7 @@ import TcType
 import TyCon
 import BasicTypes
 import Outputable
+import Platform
 import Constants
 import DynFlags
 
@@ -94,6 +95,8 @@ import DynFlags
 --		Representations
 -----------------------------------------------------------------------------
 
+-- Why are these here?
+
 addIdReps :: [Id] -> [(PrimRep, Id)]
 addIdReps ids = [(idPrimRep id, id) | id <- ids]
 
@@ -158,36 +161,9 @@ data LambdaFormInfo
 			-- of a CAF.  We want the target of the update frame to
 			-- be in the heap, so we make a black hole to hold it.
 
+                        -- XXX we can very nearly get rid of this, but
+                        -- allocDynClosure needs a LambdaFormInfo
 
--------------------------
--- An ArgDsecr describes the argument pattern of a function
-
-{-	XXX  -- imported from old ClosureInfo for now
-data ArgDescr
-  = ArgSpec		-- Fits one of the standard patterns
-	!StgHalfWord	-- RTS type identifier ARG_P, ARG_N, ...
-
-  | ArgGen	 	-- General case
-	Liveness	-- Details about the arguments
--}
-
-{-	XXX  -- imported from old ClosureInfo for now
--------------------------
--- We represent liveness bitmaps as a Bitmap (whose internal
--- representation really is a bitmap).  These are pinned onto case return
--- vectors to indicate the state of the stack for the garbage collector.
--- 
--- In the compiled program, liveness bitmaps that fit inside a single
--- word (StgWord) are stored as a single word, while larger bitmaps are
--- stored as a pointer to an array of words. 
-
-data Liveness
-  = SmallLiveness	-- Liveness info that fits in one word
-	StgWord		-- Here's the bitmap
-
-  | BigLiveness		-- Liveness info witha a multi-word bitmap
-	CLabel		-- Label for the bitmap
--}
 
 -------------------------
 -- StandardFormInfo tells whether this thunk has one of 
@@ -255,9 +231,9 @@ might_be_a_function :: Type -> Bool
 -- Return False only if we are *sure* it's a data type
 -- Look through newtypes etc as much as poss
 might_be_a_function ty
-  = case splitTyConApp_maybe (repType ty) of
-	Just (tc, _) -> not (isDataTyCon tc)
-	Nothing	     -> True
+  = case tyConAppTyCon_maybe (repType ty) of
+	Just tc -> not (isDataTyCon tc)
+	Nothing -> True
 
 -------------
 mkConLFInfo :: DataCon -> LambdaFormInfo
@@ -292,6 +268,10 @@ mkLFImported id
   where
     arity = idArity id
 
+------------
+mkLFBlackHole :: LambdaFormInfo
+mkLFBlackHole = LFBlackHole
+
 -----------------------------------------------------
 --		Dynamic pointer tagging
 -----------------------------------------------------
@@ -352,47 +332,37 @@ maybeIsLFCon _ = Nothing
 
 ------------
 isLFThunk :: LambdaFormInfo -> Bool
-isLFThunk (LFThunk _ _ _ _ _)  = True
-isLFThunk LFBlackHole          = True
+isLFThunk (LFThunk {})  = True
+isLFThunk LFBlackHole   = True
 	-- return True for a blackhole: this function is used to determine
 	-- whether to use the thunk header in SMP mode, and a blackhole
 	-- must have one.
 isLFThunk _ = False
 
+isLFReEntrant :: LambdaFormInfo -> Bool
+isLFReEntrant (LFReEntrant {}) = True
+isLFReEntrant _                = False
 
 -----------------------------------------------------------------------------
 --		Choosing SM reps
 -----------------------------------------------------------------------------
 
-chooseSMRep
-	:: Bool			-- True <=> static closure
-	-> LambdaFormInfo
-	-> WordOff -> WordOff	-- Tot wds, ptr wds
-	-> SMRep
+lfClosureType :: LambdaFormInfo -> ClosureTypeInfo
+lfClosureType (LFReEntrant _ arity _ argd) = Fun (fromIntegral arity) argd
+lfClosureType (LFCon con)                  = Constr (fromIntegral (dataConTagZ con))
+                                                      (dataConIdentity con)
+lfClosureType (LFThunk _ _ _ is_sel _) 	   = thunkClosureType is_sel
+lfClosureType _                 	   = panic "lfClosureType"
 
-chooseSMRep is_static lf_info tot_wds ptr_wds
-  = let
-	 nonptr_wds   = tot_wds - ptr_wds
-	 closure_type = getClosureType is_static ptr_wds lf_info
-    in
-    GenericRep is_static ptr_wds nonptr_wds closure_type	
+thunkClosureType :: StandardFormInfo -> ClosureTypeInfo
+thunkClosureType (SelectorThunk off) = ThunkSelector (fromIntegral off)
+thunkClosureType _                   = Thunk
 
 -- We *do* get non-updatable top-level thunks sometimes.  eg. f = g
 -- gets compiled to a jump to g (if g has non-zero arity), instead of
 -- messing around with update frames and PAPs.  We set the closure type
 -- to FUN_STATIC in this case.
 
-getClosureType :: Bool -> WordOff -> LambdaFormInfo -> ClosureType
-getClosureType is_static ptr_wds lf_info
-  = case lf_info of
-	LFCon {} | is_static && ptr_wds == 0 -> ConstrNoCaf
-		 | otherwise		     -> Constr
-  	LFReEntrant {}			     -> Fun
-	LFThunk _ _ _ (SelectorThunk {}) _   -> ThunkSelector
-	LFThunk {} 			     -> Thunk
-	_ -> panic "getClosureType"
-
-
 -----------------------------------------------------------------------------
 --		nodeMustPointToIt
 -----------------------------------------------------------------------------
@@ -555,11 +525,6 @@ getCallMethod _ _name _ LFBlackHole _n_args
 getCallMethod _ _name _ LFLetNoEscape _n_args
   = JumpToIt
 
-isStandardFormThunk :: LambdaFormInfo -> Bool
-isStandardFormThunk (LFThunk _ _ _ (SelectorThunk _) _) = True
-isStandardFormThunk (LFThunk _ _ _ (ApThunk _) _)	= True
-isStandardFormThunk _other_lf_info 			= False
-
 isKnownFun :: LambdaFormInfo -> Bool
 isKnownFun (LFReEntrant _ _ _ _) = True
 isKnownFun LFLetNoEscape	 = True
@@ -652,59 +617,50 @@ staticClosureRequired binder other_binder_info other_lf_info = True
 -}
 
 -----------------------------------------------------------------------------
---		Data types for closure information}
+--              Data types for closure information
 -----------------------------------------------------------------------------
 
 
-{- Information about a closure, from the code generator's point of view.
+{- ClosureInfo: information about a binding
 
-A ClosureInfo decribes the info pointer of a closure.  It has
-enough information 
-  a) to construct the info table itself
-  b) to allocate a closure containing that info pointer (i.e.
-	it knows the info table label)
+   We make a ClosureInfo for each let binding (both top level and not),
+   but not bindings for data constructors: for those we build a CmmInfoTable
+   directly (see mkDataConInfoTable).
 
-We make a ClosureInfo for
-  - each let binding (both top level and not)
-  - each data constructor (for its shared static and
-	dynamic info tables)
+   To a first approximation:
+       ClosureInfo = (LambdaFormInfo, CmmInfoTable)
+
+   A ClosureInfo has enough information
+     a) to construct the info table itself, and build other things
+        related to the binding (e.g. slow entry points for a function)
+     b) to allocate a closure containing that info pointer (i.e.
+   	it knows the info table label)
 -}
 
 data ClosureInfo
   = ClosureInfo {
-	closureName   :: !Name,		  -- The thing bound to this closure
-	closureLFInfo :: !LambdaFormInfo, -- NOTE: not an LFCon (see below)
-	closureSMRep  :: !SMRep,	  -- representation used by storage mgr
-	closureSRT    :: !C_SRT,	  -- What SRT applies to this closure
-	closureType   :: !Type,		  -- Type of closure (ToDo: remove)
-	closureDescr  :: !String,	  -- closure description (for profiling)
-        closureCafs   :: !CafInfo,        -- whether the closure may have CAFs
-	closureInfLcl :: Bool             -- can the info pointer be a local symbol?
+        closureName :: !Name,           -- The thing bound to this closure
+           -- we don't really need this field: it's only used in generating
+           -- code for ticky and profiling, and we could pass the information
+           -- around separately, but it doesn't do much harm to keep it here.
+
+        closureLFInfo :: !LambdaFormInfo, -- NOTE: not an LFCon
+          -- this tells us about what the closure contains: it's right-hand-side.
+
+          -- the rest is just an unpacked CmmInfoTable.
+        closureInfoLabel :: !CLabel,
+        closureSMRep     :: !SMRep,          -- representation used by storage mgr
+        closureSRT       :: !C_SRT,          -- What SRT applies to this closure
+        closureProf      :: !ProfilingInfo
     }
 
-  -- Constructor closures don't have a unique info table label (they use
-  -- the constructor's info table), and they don't have an SRT.
-  | ConInfo {
-	closureCon       :: !DataCon,
-	closureSMRep     :: !SMRep
-    }
-
-{- 	XXX temp imported from old ClosureInfo 
--- C_SRT is what StgSyn.SRT gets translated to... 
--- we add a label for the table, and expect only the 'offset/length' form
-
-data C_SRT = NoC_SRT
-	   | C_SRT !CLabel !WordOff !StgHalfWord {-bitmap or escape-}
-           deriving (Eq)
-
-instance Outputable C_SRT where
-  ppr (NoC_SRT) = ptext SLIT("_no_srt_")
-  ppr (C_SRT label off bitmap) = parens (ppr label <> comma <> ppr off <> comma <> text (show bitmap))
--}
-
-needsSRT :: C_SRT -> Bool
-needsSRT NoC_SRT       = False
-needsSRT (C_SRT _ _ _) = True
+-- | Convert from 'ClosureInfo' to 'CmmInfoTable'.
+mkCmmInfo :: ClosureInfo -> CmmInfoTable
+mkCmmInfo ClosureInfo {..}
+  = CmmInfoTable { cit_lbl  = closureInfoLabel
+                 , cit_rep  = closureSMRep
+                 , cit_prof = closureProf
+                 , cit_srt  = closureSRT }
 
 
 --------------------------------------
@@ -718,231 +674,60 @@ mkClosureInfo :: Bool		-- Is static
 	      -> C_SRT
 	      -> String		-- String descriptor
 	      -> ClosureInfo
-mkClosureInfo is_static id lf_info tot_wds ptr_wds srt_info descr
-  = ClosureInfo { closureName = name, 
-		  closureLFInfo = lf_info,
-		  closureSMRep = sm_rep, 
-		  closureSRT = srt_info,
-		  closureType = idType id,
-		  closureDescr = descr,
-                  closureCafs = idCafInfo id,
-		  closureInfLcl = isDataConWorkId id }
-		    -- Make the _info pointer for the implicit datacon worker binding
-		    -- local. The reason we can do this is that importing code always
-		    -- either uses the _closure or _con_info. By the invariants in CorePrep
-		    -- anything else gets eta expanded.
+mkClosureInfo is_static id lf_info tot_wds ptr_wds srt_info val_descr
+  = ClosureInfo { closureName      = name,
+                  closureLFInfo    = lf_info,
+                  closureInfoLabel = info_lbl,
+                  closureSMRep     = sm_rep,    -- These four fields are a
+                  closureSRT       = srt_info,  --        CmmInfoTable
+                  closureProf      = prof }     -- ---
   where
-    name   = idName id
-    sm_rep = chooseSMRep is_static lf_info tot_wds ptr_wds
+    name       = idName id
+    sm_rep     = mkHeapRep is_static ptr_wds nonptr_wds (lfClosureType lf_info)
+    prof       = mkProfilingInfo id val_descr
+    nonptr_wds = tot_wds - ptr_wds
 
-mkConInfo :: Bool	-- Is static
-	  -> DataCon	
-	  -> Int -> Int	-- Total and pointer words
-	  -> ClosureInfo
-mkConInfo is_static data_con tot_wds ptr_wds
-   = ConInfo {	closureSMRep = sm_rep,
-		closureCon = data_con }
-  where
-    sm_rep = chooseSMRep is_static (mkConLFInfo data_con) tot_wds ptr_wds
-
-
--- We need a black-hole closure info to pass to @allocDynClosure@ when we
--- want to allocate the black hole on entry to a CAF.  These are the only
--- ways to build an LFBlackHole, maintaining the invariant that it really
--- is a black hole and not something else.
-
-cafBlackHoleClosureInfo :: ClosureInfo -> ClosureInfo
-cafBlackHoleClosureInfo (ClosureInfo { closureName = nm,
-				       closureType = ty,
-				       closureCafs = cafs })
-  = ClosureInfo { closureName   = nm,
-		  closureLFInfo = LFBlackHole,
-		  closureSMRep  = BlackHoleRep,
-		  closureSRT    = NoC_SRT,
-		  closureType   = ty,
-		  closureDescr  = "", 
-		  closureCafs   = cafs,
-		  closureInfLcl = False }
-cafBlackHoleClosureInfo _ = panic "cafBlackHoleClosureInfo"
-
-
---------------------------------------
---   Extracting ClosureTypeInfo
---------------------------------------
-
--- JD: I've added the continuation arguments not for fun but because
--- I don't want to pipe the monad in here (circular module dependencies),
--- and I don't want to pull this code out of this module, which would
--- require us to expose a bunch of abstract types.
-
-closureTypeInfo ::
-  ClosureInfo -> ((ConstrDescription -> ClosureTypeInfo) -> DataCon -> CLabel -> a) ->
-  (ClosureTypeInfo -> a) -> a
-closureTypeInfo cl_info k_with_con_name k_simple
-   = case cl_info of
-	ConInfo { closureCon = con } 
-		-> k_with_con_name (ConstrInfo (ptrs, nptrs)
-			              (fromIntegral (dataConTagZ con))) con info_lbl
-		where
-		  --con_name = panic "closureTypeInfo"
-			-- Was: 
-			-- cstr <- mkByteStringCLit $ dataConIdentity con
-			-- con_name = makeRelativeRefTo info_lbl cstr
-
-    	ClosureInfo { closureName   = name,
-                      closureLFInfo = LFReEntrant _ arity _ arg_descr,
-                      closureSRT    = srt }
-		-> k_simple $ FunInfo (ptrs, nptrs)
-	                        srt 
-	                        (fromIntegral arity)
-	                        arg_descr 
-	                        (CmmLabel (mkSlowEntryLabel name (clHasCafRefs cl_info)))
-  
-  	ClosureInfo { closureLFInfo = LFThunk _ _ _ (SelectorThunk offset) _, 
-                      closureSRT    = srt }
- 		-> k_simple $ ThunkSelectorInfo (fromIntegral offset) srt
-
-    	ClosureInfo { closureLFInfo = LFThunk {}, 
-                      closureSRT    = srt }
-		-> k_simple $ ThunkInfo (ptrs, nptrs) srt
-
-        _ -> panic "unexpected lambda form in mkCmmInfo"
-  where
-    info_lbl = infoTableLabelFromCI cl_info
-    ptrs     = fromIntegral $ closurePtrsSize cl_info
-    size     = fromIntegral $ closureNonHdrSize cl_info
-    nptrs    = size - ptrs
-
---------------------------------------
---   Functions about closure *sizes*
---------------------------------------
-
-closureSize :: ClosureInfo -> WordOff
-closureSize cl_info = hdr_size + closureNonHdrSize cl_info
-  where hdr_size  | closureIsThunk cl_info = thunkHdrSize
-  		  | otherwise      	   = fixedHdrSize
-	-- All thunks use thunkHdrSize, even if they are non-updatable.
-	-- this is because we don't have separate closure types for
-	-- updatable vs. non-updatable thunks, so the GC can't tell the
-	-- difference.  If we ever have significant numbers of non-
-	-- updatable thunks, it might be worth fixing this.
-
-closureNonHdrSize :: ClosureInfo -> WordOff
-closureNonHdrSize cl_info
-  = tot_wds + computeSlopSize tot_wds cl_info
-  where
-    tot_wds = closureGoodStuffSize cl_info
-
-closureGoodStuffSize :: ClosureInfo -> WordOff
-closureGoodStuffSize cl_info
-  = let (ptrs, nonptrs) = sizes_from_SMRep (closureSMRep cl_info)
-    in	ptrs + nonptrs
-
-closurePtrsSize :: ClosureInfo -> WordOff
-closurePtrsSize cl_info
-  = let (ptrs, _) = sizes_from_SMRep (closureSMRep cl_info)
-    in	ptrs
-
--- not exported:
-sizes_from_SMRep :: SMRep -> (WordOff,WordOff)
-sizes_from_SMRep (GenericRep _ ptrs nonptrs _)   = (ptrs, nonptrs)
-sizes_from_SMRep BlackHoleRep			 = (0, 0)
-
--- Computing slop size.  WARNING: this looks dodgy --- it has deep
--- knowledge of what the storage manager does with the various
--- representations...
---
--- Slop Requirements: every thunk gets an extra padding word in the
--- header, which takes the the updated value.
-
-slopSize :: ClosureInfo -> WordOff
-slopSize cl_info = computeSlopSize payload_size cl_info
-  where payload_size = closureGoodStuffSize cl_info
-
-computeSlopSize :: WordOff -> ClosureInfo -> WordOff
-computeSlopSize payload_size cl_info
-  = max 0 (minPayloadSize smrep updatable - payload_size)
-  where
-	smrep        = closureSMRep cl_info
-	updatable    = closureNeedsUpdSpace cl_info
-
-closureNeedsUpdSpace :: ClosureInfo -> Bool
--- We leave space for an update if either (a) the closure is updatable
--- or (b) it is a static thunk.  This is because a static thunk needs
--- a static link field in a predictable place (after the slop), regardless
--- of whether it is updatable or not.
-closureNeedsUpdSpace (ClosureInfo { closureLFInfo = 
-					LFThunk TopLevel _ _ _ _ }) = True
-closureNeedsUpdSpace cl_info = closureUpdReqd cl_info
-
-minPayloadSize :: SMRep -> Bool -> WordOff
-minPayloadSize smrep updatable
-  = case smrep of
-	BlackHoleRep		 		-> min_upd_size
-	GenericRep _ _ _ _      | updatable     -> min_upd_size
-	GenericRep True _ _ _                   -> 0 -- static
-	GenericRep False _ _ _                  -> mIN_PAYLOAD_SIZE
-          --       ^^^^^___ dynamic
-  where
-   min_upd_size =
-	ASSERT(mIN_PAYLOAD_SIZE <= sIZEOF_StgSMPThunkHeader)
-	0 	-- check that we already have enough
-		-- room for mIN_SIZE_NonUpdHeapObject,
-		-- due to the extra header word in SMP
+    info_lbl = mkClosureInfoTableLabel id lf_info
 
 --------------------------------------
 --   Other functions over ClosureInfo
 --------------------------------------
 
-blackHoleOnEntry :: DynFlags -> ClosureInfo -> Bool
--- Static closures are never themselves black-holed.
--- Updatable ones will be overwritten with a CAFList cell, which points to a 
--- black hole;
--- Single-entry ones have no fvs to plug, and we trust they don't form part 
--- of a loop.
+-- Eager blackholing is normally disabled, but can be turned on with
+-- -feager-blackholing.  When it is on, we replace the info pointer of
+-- the thunk with stg_EAGER_BLACKHOLE_info on entry.
 
-blackHoleOnEntry _ ConInfo{} = False
-blackHoleOnEntry dflags (ClosureInfo { closureLFInfo = lf_info, closureSMRep = rep })
-  | isStaticRep rep
+-- If we wanted to do eager blackholing with slop filling,
+-- we'd need to do it at the *end* of a basic block, otherwise
+-- we overwrite the free variables in the thunk that we still
+-- need.  We have a patch for this from Andy Cheadle, but not
+-- incorporated yet. --SDM [6/2004]
+--
+--
+-- Previously, eager blackholing was enabled when ticky-ticky
+-- was on. But it didn't work, and it wasn't strictly necessary 
+-- to bring back minimal ticky-ticky, so now EAGER_BLACKHOLING 
+-- is unconditionally disabled. -- krc 1/2007
+
+-- Static closures are never themselves black-holed.
+
+blackHoleOnEntry :: ClosureInfo -> Bool
+blackHoleOnEntry cl_info
+  | isStaticRep (closureSMRep cl_info)
   = False	-- Never black-hole a static closure
 
   | otherwise
-  = case lf_info of
+  = case closureLFInfo cl_info of
 	LFReEntrant _ _ _ _	  -> False
 	LFLetNoEscape 		  -> False
-	LFThunk _ no_fvs updatable _ _
-	  -> if updatable
-	     then not opt_OmitBlackHoling
-	     else doingTickyProfiling dflags || not no_fvs
-                  -- the former to catch double entry,
-                  -- and the latter to plug space-leaks.  KSW/SDM 1999-04.
-
-	_other -> panic "blackHoleOnEntry"	-- Should never happen
-
-
-staticClosureNeedsLink :: ClosureInfo -> Bool
--- A static closure needs a link field to aid the GC when traversing
--- the static closure graph.  But it only needs such a field if either
--- 	a) it has an SRT
---	b) it's a constructor with one or more pointer fields
--- In case (b), the constructor's fields themselves play the role
--- of the SRT.
-staticClosureNeedsLink (ClosureInfo { closureSRT = srt })
-  = needsSRT srt
-staticClosureNeedsLink (ConInfo { closureSMRep = sm_rep, closureCon = con })
-  = not (isNullaryRepDataCon con) && not_nocaf_constr
-  where
-    not_nocaf_constr = 
-	case sm_rep of 
-	   GenericRep _ _ _ ConstrNoCaf -> False
-	   _other			-> True
+        LFThunk _ no_fvs _updatable _ _ -> not no_fvs -- to plug space-leaks.
+        _other -> panic "blackHoleOnEntry"      -- Should never happen
 
 isStaticClosure :: ClosureInfo -> Bool
 isStaticClosure cl_info = isStaticRep (closureSMRep cl_info)
 
 closureUpdReqd :: ClosureInfo -> Bool
 closureUpdReqd ClosureInfo{ closureLFInfo = lf_info } = lfUpdatable lf_info
-closureUpdReqd ConInfo{} = False
 
 lfUpdatable :: LambdaFormInfo -> Bool
 lfUpdatable (LFThunk _ _ upd _ _)  = upd
@@ -951,10 +736,6 @@ lfUpdatable LFBlackHole 	   = True
 	-- alg case with a named default... so they need to be updated.
 lfUpdatable _ = False
 
-closureIsThunk :: ClosureInfo -> Bool
-closureIsThunk ClosureInfo{ closureLFInfo = lf_info } = isLFThunk lf_info
-closureIsThunk ConInfo{} = False
-
 closureSingleEntry :: ClosureInfo -> Bool
 closureSingleEntry (ClosureInfo { closureLFInfo = LFThunk _ _ upd _ _}) = not upd
 closureSingleEntry _ = False
@@ -963,13 +744,8 @@ closureReEntrant :: ClosureInfo -> Bool
 closureReEntrant (ClosureInfo { closureLFInfo = LFReEntrant _ _ _ _ }) = True
 closureReEntrant _ = False
 
-isConstrClosure_maybe :: ClosureInfo -> Maybe DataCon
-isConstrClosure_maybe (ConInfo { closureCon = data_con }) = Just data_con
-isConstrClosure_maybe _ 				  = Nothing
-
 closureFunInfo :: ClosureInfo -> Maybe (Int, ArgDescr)
 closureFunInfo (ClosureInfo { closureLFInfo = lf_info }) = lfFunInfo lf_info
-closureFunInfo _ = Nothing
 
 lfFunInfo :: LambdaFormInfo ->  Maybe (Int, ArgDescr)
 lfFunInfo (LFReEntrant _ arity _ arg_desc)  = Just (arity, arg_desc)
@@ -977,7 +753,6 @@ lfFunInfo _                                 = Nothing
 
 funTag :: ClosureInfo -> DynTag
 funTag (ClosureInfo { closureLFInfo = lf_info }) = lfDynTag lf_info
-funTag (ConInfo {})				 = panic "funTag"
 
 isToplevClosure :: ClosureInfo -> Bool
 isToplevClosure (ClosureInfo { closureLFInfo = lf_info })
@@ -985,56 +760,57 @@ isToplevClosure (ClosureInfo { closureLFInfo = lf_info })
       LFReEntrant TopLevel _ _ _ -> True
       LFThunk TopLevel _ _ _ _   -> True
       _other			 -> False
-isToplevClosure _ = False
 
 --------------------------------------
 --   Label generation
 --------------------------------------
 
-infoTableLabelFromCI :: ClosureInfo -> CLabel
-infoTableLabelFromCI = fst . labelsFromCI
+staticClosureLabel :: Platform -> ClosureInfo -> CLabel
+staticClosureLabel platform = toClosureLbl platform .  closureInfoLabel
 
-entryLabelFromCI :: ClosureInfo -> CLabel
-entryLabelFromCI = snd . labelsFromCI
+closureRednCountsLabel :: Platform -> ClosureInfo -> CLabel
+closureRednCountsLabel platform = toRednCountsLbl platform . closureInfoLabel
 
-labelsFromCI :: ClosureInfo -> (CLabel, CLabel) -- (Info, Entry)
-labelsFromCI cl@(ClosureInfo { closureName = name,
-			       closureLFInfo = lf_info,
-			       closureInfLcl = is_lcl })
+closureSlowEntryLabel :: Platform -> ClosureInfo -> CLabel
+closureSlowEntryLabel platform = toSlowEntryLbl platform . closureInfoLabel
+
+closureLocalEntryLabel :: Platform -> ClosureInfo -> CLabel
+closureLocalEntryLabel platform
+  | tablesNextToCode = toInfoLbl  platform . closureInfoLabel
+  | otherwise        = toEntryLbl platform . closureInfoLabel
+
+mkClosureInfoTableLabel :: Id -> LambdaFormInfo -> CLabel
+mkClosureInfoTableLabel id lf_info
   = case lf_info of
-	LFBlackHole -> (mkCAFBlackHoleInfoTableLabel, mkCAFBlackHoleEntryLabel)
+        LFBlackHole -> mkCAFBlackHoleInfoTableLabel
 
-	LFThunk _ _ upd_flag (SelectorThunk offset) _ -> 
-		bothL (mkSelectorInfoLabel, mkSelectorEntryLabel) upd_flag offset
+	LFThunk _ _ upd_flag (SelectorThunk offset) _ 
+                      -> mkSelectorInfoLabel upd_flag offset
 
-	LFThunk _ _ upd_flag (ApThunk arity) _ -> 
-		bothL (mkApInfoTableLabel, mkApEntryLabel) upd_flag arity
+	LFThunk _ _ upd_flag (ApThunk arity) _ 
+                      -> mkApInfoTableLabel upd_flag arity
 
-	LFThunk{}      -> bothL std_mk_lbls name $ clHasCafRefs cl
+        LFThunk{}     -> std_mk_lbl name cafs
+        LFReEntrant{} -> std_mk_lbl name cafs
+        _other        -> panic "closureInfoTableLabel"
 
-	LFReEntrant _ _ _ _ -> bothL std_mk_lbls name $ clHasCafRefs cl
+  where 
+    name = idName id
 
-	_other -> panic "labelsFromCI"
-  where std_mk_lbls = if is_lcl then (mkLocalInfoTableLabel, mkLocalEntryLabel) else (mkInfoTableLabel, mkEntryLabel)
+    std_mk_lbl | is_local  = mkLocalInfoTableLabel
+               | otherwise = mkInfoTableLabel
 
-labelsFromCI cl@(ConInfo { closureCon = con, closureSMRep = rep })
-  | isStaticRep rep = bothL (mkStaticInfoTableLabel, mkStaticConEntryLabel)  name $ clHasCafRefs cl
-  | otherwise	    = bothL (mkConInfoTableLabel,    mkConEntryLabel)     name $ clHasCafRefs cl
-  where
-    name = dataConName con
+    cafs     = idCafInfo id
+    is_local = isDataConWorkId id
+       -- Make the _info pointer for the implicit datacon worker
+       -- binding local. The reason we can do this is that importing
+       -- code always either uses the _closure or _con_info. By the
+       -- invariants in CorePrep anything else gets eta expanded.
 
-bothL :: (a -> b -> c, a -> b -> c) -> a -> b -> (c, c)
-bothL (f, g) x y = (f x y, g x y)
-
--- ClosureInfo for a closure (as opposed to a constructor) is always local
-closureLabelFromCI :: ClosureInfo -> CLabel
-closureLabelFromCI cl@(ClosureInfo { closureName = nm }) =
-  mkLocalClosureLabel nm $ clHasCafRefs cl
-closureLabelFromCI _ = panic "closureLabelFromCI"
 
 thunkEntryLabel :: Name -> CafInfo -> StandardFormInfo -> Bool -> CLabel
--- thunkEntryLabel is a local help function, not exported.  It's used from both
--- entryLabelFromCI and getCallMethod.
+-- thunkEntryLabel is a local help function, not exported.  It's used from
+-- getCallMethod.
 thunkEntryLabel _thunk_id _ (ApThunk arity) upd_flag
   = enterApLabel upd_flag arity
 thunkEntryLabel _thunk_id _ (SelectorThunk offset) upd_flag
@@ -1057,11 +833,6 @@ enterIdLabel id c
   | tablesNextToCode = mkInfoTableLabel id c
   | otherwise        = mkEntryLabel id c
 
-enterLocalIdLabel :: Name -> CafInfo -> CLabel
-enterLocalIdLabel id c
-  | tablesNextToCode = mkLocalInfoTableLabel id c
-  | otherwise        = mkLocalEntryLabel id c
-
 
 --------------------------------------
 --   Profiling
@@ -1076,16 +847,13 @@ enterLocalIdLabel id c
 -- The type is determined from the type information stored with the @Id@
 -- in the closure info using @closureTypeDescr@.
 
-closureValDescr, closureTypeDescr :: ClosureInfo -> String
-closureValDescr (ClosureInfo {closureDescr = descr}) 
-  = descr
-closureValDescr (ConInfo {closureCon = con})
-  = occNameString (getOccName con)
-
-closureTypeDescr (ClosureInfo { closureType = ty })
-  = getTyDescription ty
-closureTypeDescr (ConInfo { closureCon = data_con })
-  = occNameString (getOccName (dataConTyCon data_con))
+mkProfilingInfo :: Id -> String -> ProfilingInfo
+mkProfilingInfo id val_descr
+  | not opt_SccProfilingOn = NoProfilingInfo
+  | otherwise = ProfilingInfo ty_descr_w8 val_descr_w8
+  where
+    ty_descr_w8  = stringToWord8s (getTyDescription (idType id))
+    val_descr_w8 = stringToWord8s val_descr
 
 getTyDescription :: Type -> String
 getTyDescription ty
@@ -1095,23 +863,58 @@ getTyDescription ty
       AppTy fun _      	     -> getTyDescription fun
       FunTy _ res      	     -> '-' : '>' : fun_result res
       TyConApp tycon _ 	     -> getOccString tycon
-      PredTy sty	     -> getPredTyDescription sty
       ForAllTy _ ty          -> getTyDescription ty
     }
   where
     fun_result (FunTy _ res) = '>' : fun_result res
     fun_result other	     = getTyDescription other
 
-getPredTyDescription :: PredType -> String
-getPredTyDescription (ClassP cl _) = getOccString cl
-getPredTyDescription (IParam ip _) = getOccString (ipNameName ip)
-getPredTyDescription (EqPred {})   = "Type equality"
-
 --------------------------------------
---   SRTs/CAFs
+--   CmmInfoTable-related things
 --------------------------------------
 
--- We need to know whether a closure may have CAFs.
-clHasCafRefs :: ClosureInfo -> CafInfo
-clHasCafRefs (ClosureInfo {closureCafs = cafs}) = cafs
-clHasCafRefs (ConInfo {}) = NoCafRefs
+mkDataConInfoTable :: DataCon -> Bool -> Int -> Int -> CmmInfoTable
+mkDataConInfoTable data_con is_static ptr_wds nonptr_wds
+ = CmmInfoTable { cit_lbl  = info_lbl
+                , cit_rep  = sm_rep
+                , cit_prof = prof
+                , cit_srt  = NoC_SRT }
+ where
+   name = dataConName data_con
+
+   info_lbl | is_static = mkStaticInfoTableLabel name NoCafRefs
+            | otherwise = mkConInfoTableLabel    name NoCafRefs
+
+   sm_rep = mkHeapRep is_static ptr_wds nonptr_wds cl_type
+
+   cl_type = Constr (fromIntegral (dataConTagZ data_con))
+                   (dataConIdentity data_con)
+
+   prof | not opt_SccProfilingOn = NoProfilingInfo
+        | otherwise              = ProfilingInfo ty_descr val_descr
+
+   ty_descr  = stringToWord8s $ occNameString $ getOccName $ dataConTyCon data_con
+   val_descr = stringToWord8s $ occNameString $ getOccName data_con
+
+-- We need a black-hole closure info to pass to @allocDynClosure@ when we
+-- want to allocate the black hole on entry to a CAF.
+
+cafBlackHoleInfoTable :: CmmInfoTable
+cafBlackHoleInfoTable
+  = CmmInfoTable { cit_lbl  = mkCAFBlackHoleInfoTableLabel
+                 , cit_rep  = blackHoleRep
+                 , cit_prof = NoProfilingInfo
+                 , cit_srt  = NoC_SRT }
+
+staticClosureNeedsLink :: CmmInfoTable -> Bool
+-- A static closure needs a link field to aid the GC when traversing
+-- the static closure graph.  But it only needs such a field if either
+-- 	a) it has an SRT
+--	b) it's a constructor with one or more pointer fields
+-- In case (b), the constructor's fields themselves play the role
+-- of the SRT.
+staticClosureNeedsLink info_tbl@CmmInfoTable{ cit_rep = smrep }
+  | isConRep smrep         = not (isStaticNoCafCon smrep)
+  | otherwise              = needsSRT (cit_srt info_tbl)
+staticClosureNeedsLink _ = False
+
