@@ -1,4 +1,11 @@
 
+{-# OPTIONS -fno-warn-tabs #-}
+-- The above warning supression flag is a temporary kludge.
+-- While working on this module you are encouraged to remove it and
+-- detab the module (please do the detabbing in a separate patch). See
+--     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#TabsvsSpaces
+-- for details
+
 module RegAlloc.Graph.SpillCost (
 	SpillCostRecord,
 	plusSpillCostRecord,
@@ -65,7 +72,7 @@ plusSpillCostRecord (r1, a1, b1, c1) (r2, a2, b2, c2)
 --
 slurpSpillCostInfo :: (PlatformOutputable instr, Instruction instr)
                    => Platform
-                   -> LiveCmmTop statics instr
+                   -> LiveCmmDecl statics instr
                    -> SpillCostInfo
 
 slurpSpillCostInfo platform cmm
