@@ -1,3 +1,5 @@
+{-# LANGUAGE StandaloneDeriving, FlexibleInstances, UndecidableInstances, IncoherentInstances #-}
+{-# OPTIONS_GHC -fno-warn-orphans #-}
 module Main (main) where
 
 import Test.HUnit
@@ -6,9 +8,13 @@ import DynFlags (defaultDynFlags)
 import Haddock.Lex (tokenise)
 import Haddock.Parse (parseParas)
 import Haddock.Types
+import Outputable
 
-instance Show RdrName where
-  show x = "RdrName"
+instance Outputable a => Show a where
+  show = showSDoc . ppr
+
+deriving instance Show a => Show (Doc a)
+deriving instance Eq a =>Eq (Doc a)
 
 data ParseTest = ParseTest {
     input   :: String
@@ -23,12 +29,12 @@ tests = [
     }
 
   , ParseTest {
-      input  = "foobar\n\nghci> fib 10\n55"
+      input  = "foobar\n\n>>> fib 10\n55"
     , result = Just $ DocAppend (DocParagraph $ DocString "foobar\n") (DocExamples $ [Example "fib 10" ["55"]])
     }
 
   , ParseTest {
-      input  = "foobar\nghci> fib 10\n55"
+      input  = "foobar\n>>> fib 10\n55"
     , result = Nothing -- parse error
     }
 
@@ -41,16 +47,23 @@ tests = [
       input  = "foobar\n> some code"
     , result = Nothing -- parse error
     }
+
+  -- test <BLANKLINE> support
+  , ParseTest {
+      input  = ">>> putFooBar\nfoo\n<BLANKLINE>\nbar"
+    , result = Just $ DocExamples $ [Example "putFooBar" ["foo","","bar"]]
+    }
   ]
 
 
+main :: IO ()
 main = do
   _ <- runTestTT $ TestList $ map toTestCase tests
   return ();
   where
 
     toTestCase :: ParseTest -> Test
-    toTestCase (ParseTest input result) = TestCase $ assertEqual input (parse input) result
+    toTestCase (ParseTest s r) = TestCase $ assertEqual s r (parse s)
 
     parse :: String -> Maybe (Doc RdrName)
-    parse input = parseParas $ tokenise defaultDynFlags input (0,0)
+    parse s = parseParas $ tokenise (defaultDynFlags undefined) s (0,0)
